
-- Code in this file was taken or adapted from LibStatLogic-1.2 which is supposedly licensed under
-- LGPLv2.1 but does not have the accompanying notices. Thus, the code in this file is likewise
-- licensed under LGPLv2.1.

ItemStats = {}

-- from StatLogic
local function HasTitanGrip(talentGroup)
  if playerClass == "WARRIOR" then
    local _, _, _, _, r = GetTalentInfo(2, 20, nil, nil, talentGroup)
    return r > 0
  end
end

-- from StatLogic
local function diffTable(op1, op2)
	if type(op2) == "table" then
	  for k, v in pairs(op2) do
	    if type(v) == "number" then
	      op1[k] = (op1[k] or 0) - v
	      if op1[k] == 0 then
	        op1[k] = nil
	      end
	    end
	  end
	end
	return op1
end

-- from StatLogic
-- deletes the contents of a table, then returns itself
local function clearTable(t)
  if t then
    for k in pairs(t) do
      if type(t[k]) == "table" then
        del(t[k]) -- child tables get put into the pool
      else
        t[k] = nil
      end
    end
    setmetatable(t, nil)
  end
  return t
end

-- copyTable from StatLogic
local function copyTable(to, from)
  if not clearTable(to) then
    to = {}
  end
  for k,v in pairs(from) do
    if type(k) == "table" then
      k = copyTable(new(), k)
    end
    if type(v) == "table" then
      v = copyTable(new(), v)
    end
    to[k] = v
  end
  setmetatable(to, getmetatable(from))
  return to
end

-- from StatLogic
function ItemStats.RemoveEnchant(link)
  -- check link
  if not strfind(link, "item:%d+:%d+:%d+:%d+:%d+:%d+:%-?%d+:%-?%d+") then
    return link
  end
  local linkType, itemId, enchantId, rest = strsplit(":", link, 4)
  return strjoin(":", linkType, itemId, 0, rest)
end

function ItemStats.GetStats(item, itemTable)
	if not item then
		return
	end
	
	if type(item) == "table" and type(item.GetItem) == "function" then
		_, item = item:GetItem()
	end
	
	wipe(itemTable)
	
	local blizzardTable = GetItemStats(item, itemTable)
	itemTable.link = item
	_, _, _, _, _, itemTable.itemType, _, _, itemTable.itemSlot = GetItemInfo(itemTable.link)
end

-- Adapted from StatLogic
function ItemStats.GetDiff(item, linkDiff1, linkDiff2)
	if type(item) == "table" and type(item.GetItem) == "function" then
		_, item = item:GetItem()
	end
	
	local diff1, diff2
	itemSum = {}
	
	ItemStats.GetStats(item, itemSum)
		
	if itemSum.itemSlot == "INVTYPE_2HWEAPON" and not HasTitanGrip() then
	    local equippedSum1 = {}
	    local equippedSum2 = {}
    
	    -- Get main hand item sum
	    if linkDiff1 and linkDiff1 ~= "NOITEM" then
	       	ItemStats.GetStats(linkDiff1, equippedSum1)
	    end
	    
	    -- Get off hand item sum
	    if linkDiff2 and linkDiff2 ~= "NOITEM" then
	      	ItemStats.GetStats(linkDiff2, equippedSum2)
	    end
    
    	-- Calculate diff
    	diff1 = {}
	    copyTable(diff1, itemSum)
    	diffTable(diff1, equippedSum1)
    	diffTable(diff1, equippedSum2)
  	else
    	local equippedSum = {}
    
    	-- Get equipped item 1 sum
    	if linkDiff1 and linkDiff1 ~= "NOITEM" then
      		ItemStats.GetStats(linkDiff1, equippedSum)
      	end
    	-- Calculate item 1 diff
    	diff1 = {}
    	copyTable(diff1, itemSum)
    	diffTable(diff1, equippedSum)
    
	    -- Check if item has a second equip slot
    	if linkDiff2 then -- If so
	      	-- Get equipped item 2 sum
	      	if linkDiff2 ~= "NOITEM" then
	        	ItemStats.GetStats(linkDiff2, equippedSum)
	      	end
	      
	      	-- Calculate item 2 diff
	      	diff2 = {}
	      	copyTable(diff2, itemSum)
	    	diffTable(diff2, equippedSum)
    	end
  	end
  	
  	return diff1, diff2
end

local getSlotID = {
  INVTYPE_AMMO           = 0,
  INVTYPE_GUNPROJECTILE  = 0,
  INVTYPE_BOWPROJECTILE  = 0,
  INVTYPE_HEAD           = 1,
  INVTYPE_NECK           = 2,
  INVTYPE_SHOULDER       = 3,
  INVTYPE_BODY           = 4,
  INVTYPE_CHEST          = 5,
  INVTYPE_ROBE           = 5,
  INVTYPE_WAIST          = 6,
  INVTYPE_LEGS           = 7,
  INVTYPE_FEET           = 8,
  INVTYPE_WRIST          = 9,
  INVTYPE_HAND           = 10,
  INVTYPE_FINGER         = {11,12},
  INVTYPE_TRINKET        = {13,14},
  INVTYPE_CLOAK          = 15,
  INVTYPE_WEAPON         = {16,17},
  INVTYPE_2HWEAPON       = 16+17,
  INVTYPE_WEAPONMAINHAND = 16,
  INVTYPE_WEAPONOFFHAND  = 17,
  INVTYPE_SHIELD         = 17,
  INVTYPE_HOLDABLE       = 17,
  INVTYPE_RANGED         = 18,
  INVTYPE_RANGEDRIGHT    = 18,
  INVTYPE_RELIC          = 18,
  INVTYPE_GUN            = 18,
  INVTYPE_CROSSBOW       = 18,
  INVTYPE_WAND           = 18,
  INVTYPE_THROWN         = 18,
  INVTYPE_TABARD         = 19,
}

-- adapted from StatLogic
function ItemStats.GetDiffID(item)
  local _, name, itemType, link, linkDiff1, linkDiff2
  -- Check item
  if (type(item) == "string") or (type(item) == "number") then
  elseif type(item) == "table" and type(item.GetItem) == "function" then
    -- Get the link
    _, item = item:GetItem()
    if type(item) ~= "string" then return end
  else
    return
  end
  -- Check if item is in local cache
  name, link, _, _, _, _, _, _, itemType = GetItemInfo(item)
  if not name then return end
  -- Get equip location slot id for use in GetInventoryItemLink
  local slotID = getSlotID[itemType]
  -- Don't do bags
  if not slotID then return end

  -- 1h weapon, check if player can dual wield, check for 2h equipped
  if itemType == "INVTYPE_WEAPON" then
    linkDiff1 = GetInventoryItemLink("player", 16) or "NOITEM"
    -- If player can Dual Wield, calculate offhand difference
    if IsUsableSpell(GetSpellInfo(674)) then		-- ["Dual Wield"]
      local _, _, _, _, _, _, _, _, eqItemType = GetItemInfo(linkDiff1)
      -- If 2h is equipped, copy diff1 to diff2
      if eqItemType == "INVTYPE_2HWEAPON" and not HasTitanGrip() then
        linkDiff2 = linkDiff1
      else
        linkDiff2 = GetInventoryItemLink("player", 17) or "NOITEM"
      end
    end
  -- Ring or trinket
  elseif type(slotID) == "table" then
    -- Get slot link
    linkDiff1 = GetInventoryItemLink("player", slotID[1]) or "NOITEM"
    linkDiff2 = GetInventoryItemLink("player", slotID[2]) or "NOITEM"
  -- 2h weapon, so we calculate the difference with equipped main hand and off hand
  elseif itemType == "INVTYPE_2HWEAPON" then
    linkDiff1 = GetInventoryItemLink("player", 16) or "NOITEM"
    linkDiff2= GetInventoryItemLink("player", 17) or "NOITEM"
  -- Off hand slot, check if we have 2h equipped
  elseif slotID == 17 then
    linkDiff1 = GetInventoryItemLink("player", 16) or "NOITEM"
    -- If 2h is equipped
    local _, _, _, _, _, _, _, _, eqItemType = GetItemInfo(linkDiff1)
    if eqItemType ~= "INVTYPE_2HWEAPON" then
      linkDiff1 = GetInventoryItemLink("player", 17) or "NOITEM"
    end
  -- Single slot item
  else
    linkDiff1 = GetInventoryItemLink("player", slotID) or "NOITEM"
  end

  linkDiff1 = ItemStats.RemoveEnchant(linkDiff1)
  if linkDiff2 then
    linkDiff2 = ItemStats.RemoveEnchant(linkDiff2)
  end
  
  return linkDiff1, linkDiff2
end
