--[[
	main.lua
		Calculate and display item scores on tooltips.
--]]

ItemScore = LibStub("AceAddon-3.0"):NewAddon("ItemScore", 'AceConsole-3.0')

-- Libraries --
local TipHooker = LibStub("LibTipHooker-1.1")

-- Local data --

--[[ ItemScore functions ]]--

function ItemScore:CheckDB()
	if not self.db.char.scoreSets then
		-- create default score set
		self.db.char.scoreSets = {}
		
		self.db.char.scoreSets["Default"] = self:CreateEmptyScoreSet("Default")
	else
		for i in pairs(self.db.char.scoreSets) do
			self:MigrateSet(self.db.char.scoreSets[i].info)
		end
	end
end

function ItemScore:CreateEmptyScoreSet(name)
	local ret = {}
	
	ret.name = name
	ret.display = 1
	ret.r = 0
	ret.g = 0
	ret.b = 1 -- default to blue
	ret.pattern = "%n: %a" -- Default
	ret.precision = 2
	ret.info = {}
	
	return ret
end

local function moveKey(table, key, newKey)
	if table[key] then
		table[newKey] = table[key]
		table[key] = nil
	end
end

function ItemScore:MigrateSet(scoreSet)
	moveKey(scoreSet, 'SPELL_DMG', 'SPELL_POWER')
	moveKey(scoreSet, 'HEAL', 'SPELL_POWER')
	moveKey(scoreSet, 'RANGED_AP', 'AP')
	moveKey(scoreSet, 'FERAL_AP', 'AP')
end

function ItemScore:InitStats()
	-- store list of StatLogic stat names to provide scores for
	self.supportedStats = 
	{
		["EMPTY_SOCKET_RED"] = 'EMPTY_SOCKET_RED',
		["EMPTY_SOCKET_YELLOW"] = 'EMPTY_SOCKET_YELLOW',
		["EMPTY_SOCKET_BLUE"] = 'EMPTY_SOCKET_BLUE',
		["EMPTY_SOCKET_META"] = 'EMPTY_SOCKET_META',

		["STR"] = 'ITEM_MOD_STRENGTH_SHORT',
		["AGI"] = 'ITEM_MOD_AGILITY_SHORT',
		["STA"] = 'ITEM_MOD_STAMINA_SHORT',
		["INT"] = 'ITEM_MOD_INTELLECT_SHORT',
		["SPI"] = 'ITEM_MOD_SPIRIT_SHORT',
		["ARMOR"] = 'RESISTANCE0_NAME',
		--["ARMOR_BONUS"] = true,

		["FIRE_RES"] = 'RESISTANCE2_NAME',
		["NATURE_RES"] = 'RESISTANCE3_NAME',
		["FROST_RES"] = 'RESISTANCE4_NAME',
		["SHADOW_RES"] = 'RESISTANCE5_NAME',
		["ARCANE_RES"] = 'RESISTANCE6_NAME',

		["BLOCK_VALUE"] = 'ITEM_MOD_BLOCK_VALUE_SHORT',

		["AP"] = 'ITEM_MOD_ATTACK_POWER_SHORT',
		
		["SPELL_POWER"] = 'ITEM_MOD_SPELL_POWER_SHORT',
		
		["SPELLPEN"] = 'ITEM_MOD_SPELL_PENETRATION_SHORT',

		["HEALTH"] = 'ITEM_MOD_HEALTH_SHORT',
		["MANA"] = 'ITEM_MOD_MANA_SHORT',
		["HEALTH_REG"] = 'ITEM_MOD_HEALTH_REGENERATION_SHORT',
		["MANA_REG"] = 'ITEM_MOD_MANA_REGENERATION_SHORT',

		--["MAX_DAMAGE"] = true,
		["DPS"] = 'ITEM_MOD_DAMAGE_PER_SECOND_SHORT',

		["DEFENSE_RATING"] = 'ITEM_MOD_DEFENSE_SKILL_RATING_SHORT',
		["DODGE_RATING"] = 'ITEM_MOD_DODGE_RATING_SHORT',
		["PARRY_RATING"] = 'ITEM_MOD_PARRY_RATING_SHORT',
		["BLOCK_RATING"] = 'ITEM_MOD_BLOCK_RATING_SHORT',
		["MELEE_HIT_RATING"] = 'ITEM_MOD_HIT_RATING_SHORT',
		--["RANGED_HIT_RATING"] = true,
		--["SPELL_HIT_RATING"] = true,
		["MELEE_CRIT_RATING"] = 'ITEM_MOD_CRIT_RATING_SHORT',
		--["RANGED_CRIT_RATING"] = true,
		--["SPELL_CRIT_RATING"] = true,
		["RESILIENCE_RATING"] = 'ITEM_MOD_RESILIENCE_RATING_SHORT',
		["MELEE_HASTE_RATING"] = 'ITEM_MOD_HASTE_RATING_SHORT',
		--["RANGED_HASTE_RATING"] = true,
		--["SPELL_HASTE_RATING"] = true,
		["EXPERTISE_RATING"] = 'ITEM_MOD_EXPERTISE_RATING_SHORT',
		["ARMOR_PENETRATION_RATING"] = 'ITEM_MOD_ARMOR_PENETRATION_RATING_SHORT',
		["MASTERY_RATING"] = 'ITEM_MOD_MASTERY_RATING_SHORT'
	}
	
	-- now set the values of each index to readable values from stat logic
	--for i in pairs(self.supportedStats) do
	--	self.supportedStats[i] = StatLogic:GetStatNameFromID(i)
	--end
	
	-- now make reverse table
	self.blizzardToSupported = {}
	for supportedVal, blizzardVal in pairs(self.supportedStats) do
		self.blizzardToSupported[blizzardVal] = supportedVal
	end
end

function ItemScore:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("ItemScoreDB")
	self:InitStats()
	self:CheckDB()
	
	self.options = {}
	self:InitOptions()
	self:RegisterSlashCommands()
	
	LibStub("AceConfig-3.0"):RegisterOptionsTable("ItemScore", self.options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("ItemScore", "ItemScore")
end

function ItemScore:OnEnable()
    -- Called when the addon is enabled
    
    -- Hook item tooltips
	TipHooker:Hook(self.ProcessTooltip, "item")
end

function ItemScore:OnDisable()
    -- Called when the addon is disabled
    
    -- Unhook item tooltips
	TipHooker:Unhook(self.ProcessTooltip, "item")
end

--[[ Local Functions ]]--

--[[
	The Code
--]]

function ItemScore:ConvertBlizzardStatsToSupportedStats(itemTable, supportedTable)
	wipe(supportedTable)
	for blizzardKey, statVal in pairs(itemTable) do
		local supportedKey = self.blizzardToSupported[blizzardKey]
		if supportedKey then
			supportedTable[supportedKey] = statVal
		--else
		--	self:Print("Unsupported stat: " .. blizzardKey)
		end
	end	
end

function ItemScore:GetTotalCurrentStats(currentStats, stat)
	-- calculate total stats of currently worn equipment
	if currentStats[stat] then return end
	
	currentStats[stat] = 0
	
	for i = 0, 19 do
		local link = GetInventoryItemLink("player", i)
		local blizzardItemTable = {}
		local itemTable = {}
		ItemStats.GetStats(link, blizzardItemTable)
		self:ConvertBlizzardStatsToSupportedStats(blizzardItemTable, itemTable)
		if itemTable[stat] then
			currentStats[stat] = currentStats[stat] + itemTable[stat]
		end
	end
end

local function getCappedValue(cap, currentTotal, value)
	-- returns the amount of value counted such that currentTotal + value <= cap
	local statLeft = math.max(0, cap - currentTotal)
	
	return math.min(value, statLeft)
end

function ItemScore:CalculateItemScore(blizzardItemTable, blizzardDiff1, blizzardDiff2, scoreSet, currentStats)
	local score = 0
	local score2 = 0
	
	local equippedScore = 0
	local equippedScore2 = 0
	
	-- convert Blizzard stats to supported stats
	local itemTable = {}
	local diff1, diff2
	
	ItemScore:ConvertBlizzardStatsToSupportedStats(blizzardItemTable, itemTable)
	
	if blizzardDiff1 then
		diff1 = {}
		ItemScore:ConvertBlizzardStatsToSupportedStats(blizzardDiff1, diff1)
	end
	
	if blizzardDiff2 then
		diff2 = {}
		ItemScore:ConvertBlizzardStatsToSupportedStats(blizzardDiff2, diff2)
	end
	
	for i in pairs(self.supportedStats) do
		local thisScore = 0
		local thisCap = -1
		
		if scoreSet.info[i] then
			thisScore = scoreSet.info[i].score
			thisCap = scoreSet.info[i].cap
		elseif scoreSet.parent and ItemScore.db.char.scoreSets[scoreSet.parent] and ItemScore.db.char.scoreSets[scoreSet.parent].info[i] then
			thisScore = ItemScore.db.char.scoreSets[scoreSet.parent].info[i].score
			thisCap = ItemScore.db.char.scoreSets[scoreSet.parent].info[i].cap
		end 
		
		if thisScore ~= 0 then
			if not itemTable[i] then
				itemTable[i] = 0
			end
			-- calculate current stats for those with caps
			
			-- calculate for current equipment too
			local current1Val = 0
			local current2Val = 0
			
			if diff1 then
				if not diff1[i] then
					-- no difference
					diff1[i] = 0
				end
				
				current1Val = itemTable[i] - diff1[i]
			else
				current1Val = 0	
			end
			
			current2Val = current1Val
			
			if diff2 then
				if not diff2[i] then
					-- no difference
					diff2[i] = 0
				end
				
				current2Val = itemTable[i] - diff2[i]
			end
			
			if thisCap ~= -1 then
				-- calculate capped value
				-- get current stat (as needed)
				self:GetTotalCurrentStats(currentStats, i)
				
				-- determine current value(s) of currently equipped items
				local add = 0
				local equipAdd = 0
				
				-- get capped value
				add = getCappedValue(thisCap, currentStats[i] - current1Val, itemTable[i]) * thisScore
				score = score + add
				
				-- do for equipped item as well
				equipAdd = getCappedValue(thisCap, currentStats[i] - current1Val, current1Val) * thisScore
				equippedScore = equippedScore + equipAdd
			
				if current2Val ~= current1Val then
					add = getCappedValue(thisCap, currentStats[i] - current2Val, itemTable[i]) * thisScore
					
					-- do for equipped item as well
					equipAdd = getCappedValue(thisCap, currentStats[i] - current2Val, current2Val) * thisScore
				end
				
				score2 = score2 + add
				equippedScore2 = equippedScore2 + equipAdd

			else
			    local add = thisScore * itemTable[i]
				score = score + add 
				score2 = score2 + add
				equippedScore = equippedScore + (thisScore * current1Val)
				equippedScore2 = equippedScore2 + (thisScore * current2Val)
			end
		end   
	end
	
	return score, score2, equippedScore, equippedScore2
end

function ItemScore:MakeTooltipText(scoreSet, score, score2, equippedScore, equippedScore2)
	
	local scoreText = string.format("%." .. scoreSet.precision .. "f", score)
			
	if score ~= score2 then
		scoreText = scoreText .. "/" .. string.format("%." .. scoreSet.precision .. "f", score2)
	end
	
	local diff = score - equippedScore
	local sign = ""
	local signColor = "|cFFFF0000"
	
	if diff >= 0 then
		sign = "+"
		signColor = "|cFF00FF00"
	end
	
	local equippedDiffText = signColor .. sign .. string.format("%." .. scoreSet.precision .. "f", diff) .. "|r"
	
	if score ~= score2 or equippedScore ~= equippedScore2 then
		diff = score2 - equippedScore2
		if diff >= 0 then
			sign = "+"
			signColor = "|cFF00FF00"
		else
			sign = ""
			signColor = "|cFFFF0000"
		end
			
		equippedDiffText = equippedDiffText .. "/" .. signColor ..  sign .. string.format("%." .. scoreSet.precision .. "f", diff) .. "|r"
	end
			
	local ret = scoreSet.pattern
	
	ret = string.gsub(ret, "%%n", scoreSet.name)
	ret = string.gsub(ret, "%%a", scoreText)
	ret = string.gsub(ret, "%%d", equippedDiffText)
	
	return ret
end

function ItemScore.ProcessTooltip(tooltip, name, link, ...)
	local currentStats = {}
	local itemTable = {}	
	
	-- get the item stats
	ItemStats.GetStats(tooltip, itemTable)
	
	-- check if we're a gem
	local isGem = nil
	if itemTable.itemType == "Gem" then
		isGem = true
	end
	
	-- if we're not equippable or we are tabard, bag or quiver, then don't show scores
	if not isGem and itemTable and 
		(not itemTable.itemType or itemTable.itemType == "" or
		not itemTable.itemSlot or itemTable.itemSlot == "" or 
		itemTable.itemSlot == "INVTYPE_TABARD" or itemTable.itemSlot == "INVTYPE_BAG" or
		itemTable.itemSlot == "INVTYPE_QUIVER") then
		return
	end
	
	-- by default, we'll score and compare items in their unenchanted state
	local unchantedLink = ItemStats.RemoveEnchant(itemTable.link)
		
	if unchantedLink ~= itemTable.link then
		itemTable.link = unchantedLink
		ItemStats.GetStats(itemTable.link, itemTable)
	end
	
	-- get difference in stat between equipped and this equipment
	local linkDiff1, linkDiff2 = ItemStats.GetDiffID(itemTable.link)
    
    local equipDiff1, equipDiff2 = ItemStats.GetDiff(tooltip, linkDiff1, linkDiff2)
	
	for i in pairs(ItemScore.db.char.scoreSets) do
		if ItemScore.db.char.scoreSets[i].display then
			local setName = ItemScore.db.char.scoreSets[i].name
			
			-- calculate the score
			local score, score2, equippedScore, equippedScore2 = ItemScore:CalculateItemScore(itemTable, equipDiff1, equipDiff2, ItemScore.db.char.scoreSets[i], currentStats)
			
			-- display the score
			tooltip:AddLine(ItemScore:MakeTooltipText(ItemScore.db.char.scoreSets[i], score, score2, equippedScore, equippedScore2), ItemScore.db.char.scoreSets[i].r, ItemScore.db.char.scoreSets[i].g, ItemScore.db.char.scoreSets[i].b)
			tooltip:Show()
		end
	end
end

function ItemScore:PrintTotalScores()
	local currentStats = {}
	self:Print("Total scores:")
	
	for j in pairs(ItemScore.db.char.scoreSets) do
		local totalScore = 0
		local scoreSet = ItemScore.db.char.scoreSets[j]
		
		for i in pairs(self.supportedStats) do
			local thisScore = 0
			local thisCap = -1
			
			if scoreSet.info[i] then
				thisScore = scoreSet.info[i].score
				thisCap = scoreSet.info[i].cap
			elseif scoreSet.parent and ItemScore.db.char.scoreSets[scoreSet.parent] and ItemScore.db.char.scoreSets[scoreSet.parent].info[i] then
				thisScore = ItemScore.db.char.scoreSets[scoreSet.parent].info[i].score
				thisCap = ItemScore.db.char.scoreSets[scoreSet.parent].info[i].cap
			end 
			if thisScore ~= 0 then
				self:GetTotalCurrentStats(currentStats, i)
				
				if thisCap ~= -1 then
					totalScore = totalScore + (thisScore * getCappedValue(thisCap, 0, currentStats[i]))
				else
					totalScore = totalScore + (thisScore * currentStats[i])
				end
			end
		end
		
		self:Print(scoreSet.name .. ": " .. string.format("%." .. scoreSet.precision .. "f", totalScore))
	end
end

function ItemScore:OnCmd(args)
	local cmd = string.split(' ', args):lower() or args:lower()
	local restOfString = string.gsub(args, cmd .. ' ', '')

	self:PrintTotalScores()
end
